########################################################################

'''
A place to define vaccination strategies.

Author: Kuo Chun-Lin
Date  : 2020.06.06 - 2020.08.11
'''

########################################################################

import copy
import numpy as np
import networkx as nx

from math import ceil
from itertools import combinations
from cdlib import algorithms as algo

from epidemic import EpiModel
from utils.partition import *

########################################################################


class VacciModel(EpiModel):
    """docstring for VacciModel"""

    def __init__(self, G, eq, num_state, params,
                 data_dir=None, vacci_strategy=None,
                 state_colors=['blue', 'red', 'green', 'orange']):
        super(VacciModel, self).__init__(
            G, eq, num_state, params, data_dir, state_colors)

        if vacci_strategy == 'random':
            self.vacci_func = self.random_vaccination
        elif vacci_strategy == 'target':
            self.vacci_func = self.degree_based_target
        elif vacci_strategy == 'acquaintance':
            self.vacci_func = self.acquaintance
        elif vacci_strategy == 'hhi':
            self.vacci_func = self.hhi

    # initialize a certain state to probability = 1.
    def initialize_nodes(self, idx_nodes, state=0):
        # Listing the total states included in the epidemic model and ...
        # ... loop through each state.
        for s in np.arange(self.num_state):
            if s == state:
                self.model.initial[s * self.N + idx_nodes] = 1
            else:
                self.model.initial[s * self.N + idx_nodes] = 0

    def community_vaccination(self, A, param=[8, 0.5]):
        # Use an algo to detect communities.
        communities = algo.louvain(nx.Graph(A)).communities
        # Get the nodes with highest "out_community" links.
        opt1 = target_nodes(communities, param[0], out_community=True)[0]
        opt2 = self.select_nodes(A, state=0, nei_thres=0)
        opt = np.intersect1d(opt1, opt2)

        idx = np.random.choice(opt, ceil(len(opt) * param[1]), replace=False)
        A[idx, :] = 0
        A[:, idx] = 0

        self.initialize_nodes(idx, state=0)
        self.vacci_num = len(idx)
        return A

    def random_vaccination(self, A, ratio):
        # opt = self.select_nodes(A, state=0, nei_thres=0)
        opt = self.select_neighbors(A, state=1, nei_state=0, nei_thres=0)

        # Randomly pick up certain selected nodes.
        idx = np.random.choice(opt, ceil(len(opt) * ratio), replace=False)
        A[idx, :] = 0
        A[:, idx] = 0

        # When the "s" nodes are separated, set the prob of "s" to 1 again.
        self.initialize_nodes(idx, state=0)
        self.vacci_num = len(idx)
        return A

    def degree_based_target(self, A, param):    # param = (kc, ratio)
        opt = self.select_nodes(A, state=0, nei_thres=param[0])
        # opt = self.select_neighbors(A, state=1, nei_state=0, nei_thres=param[0])

        idx = np.random.choice(opt, ceil(len(opt) * param[1]), replace=False)
        A[idx, :] = 0
        A[:, idx] = 0

        # When the "s" nodes are separated, set the prob of "s" to 1 again.
        self.initialize_nodes(idx, state=0)
        self.vacci_num = len(idx)
        return A

    def acquaintance(self, A, ratio):
        opt = self.select_nodes(A, state=0, nei_thres=0)
        # opt = self.select_neighbors(A, state=1, nei_state=0, nei_thres=0)

        # Randomly pick up certain selected nodes.
        idx = np.random.choice(opt, ceil(len(opt) * ratio), replace=False)
        # Pick up oonly those nodes with neighbors.
        neighbors = np.sum(A[idx], axis=0) > 0
        A[neighbors, :] = 0
        A[:, neighbors] = 0

        # When the "s" nodes are separated, set the prob of "s" to 1 again.
        self.initialize_nodes(idx, state=0)
        self.vacci_num = np.sum(neighbors)
        return A

    def hhi(self, A, ratio):
        opt = self.select_nodes(A, state=0, nei_thres=0)
        # opt = self.select_neighbors(A, state=1, nei_state=0, nei_thres=0)

        # Determine all possible combination of the selected nodes.
        subset = np.array(list(combinations(opt, ceil(len(opt) * ratio))))
        # Create an array to save the HHI value.
        H_list = np.zeros(len(subset))

        for i, idx_nodes in enumerate(subset):
            # Try each combination by changing its A matrix.
            A_temp = copy.deepcopy(A)
            A_temp[idx_nodes.astype(np.int), :] = 0
            A_temp[:, idx_nodes.astype(np.int)] = 0
            # Get the corresponding graph using the A matrix.
            G_temp = self.set_graph(self.state_list,
                                    self.color_list,
                                    G=nx.Graph(A_temp))
            # Find the total nodes of each subgraphs.
            subgraphs = list(nx.connected_components(G_temp))
            # Count how many nodes there are for each subgraph.
            node_num_list = np.array([len(z) for z in subgraphs])
            # Calculate HHI value using the following formula.
            H = 1 - (np.sum(node_num_list) ** 2) / (len(opt) ** 2)
            # Save the value to the array.
            H_list[i] = H

        # Find the one combination with minimum HHI value.
        idx = subset[np.argmin(H_list)].astype(np.int)
        # Change the A matrix according to the selected nodes.
        A[idx, :] = 0
        A[:, idx] = 0

        # When the "s" nodes are separated, set the prob of "s" to 1 again.
        self.initialize_nodes(idx, state=0)
        self.vacci_num = len(idx)
        return A


########################################################################


if __name__ == '__main__':

    from utils.plot import draw_probs
    from equations import SI, SIS, SIR, SIRV

    G = nx.watts_strogatz_graph(40, 5, 0.4)
    days = np.arange(0, 10, 0.1)
    colors = ['blue', 'red', 'green', 'orange']

    epi = VacciModel(G, SIR, num_state=3, params=[4, 2, 0.3, 0.1],
                     vacci_strategy='random', state_colors=colors)

    probs = epi.simulate(days)
    # probs = epi.reconstruct(days, 20, param=0.5)

    epi.set_propagate(0, 1, neighbor=1, update_state=False)
    epi.set_propagate(1, 2, neighbor=None, update_state=False)

    # status, _ = epi.propagation()
    status, probs = epi.propagation(reconstruct_index=[2, 4, 7, 20],
                                    reconstruct_param=0.5)

    epi.visualize(status, np.arange(16), figsize=(15, 15), n_row=4, n_col=4)
    draw_probs(probs, colors)



########################################################################

'''

Reference:
+ scipy求解微分方程: https://www.twblogs.net/a/5b7afb882b7177539c248a16

Author: Kuo Chun-Lin
Date  : 2020.05.01
'''

########################################################################

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from scipy.integrate import odeint

########################################################################


class MacroODE(object):
    N = None
    initial = None

    def derivative(self, z, t):
        assert self.N is not None, 'N is not defined.'
        assert self.initial is not None, 'States are not initiated.'
        raise NotImplementedError()


class SI(MacroODE):
    def __init__(self, N, I0, beta):
        self.N = N
        self.I0 = I0
        self.S0 = N - I0
        self.initial = (self.S0, I0)

        # Beta: transmission / infection Rate
        self.beta = beta

    def derivative(self, y, t):
        S, I = y
        dSdt = -self.beta * S * I / self.N
        dIdt = self.beta * S * I / self.N
        return dSdt, dIdt


class SIS(MacroODE):
    def __init__(self, N, I0, beta, gamma):
        self.N = N
        self.I0 = I0
        self.S0 = N - I0
        self.initial = (self.S0, I0)

        self.beta = beta
        self.gamma = gamma
        self.reproduction_num = beta / gamma    # Definition of "R_0".

    def derivative(self, y, t):
        S, I = y
        dSdt = -self.beta * S * I / self.N + self.gamma * I
        dIdt = self.beta * S * I / self.N - self.gamma * I
        return dSdt, dIdt


class SIR(MacroODE):
    """docstring for SIR"""

    def __init__(self, N, I0, R0, beta, gamma):
        self.N = N
        self.I0 = I0
        self.R0 = R0
        self.S0 = N - I0 - R0
        self.initial = (self.S0, I0, R0)

        self.beta = beta
        self.gamma = gamma
        self.reproduction_num = beta / gamma    # Definition of "R_0".

    def derivative(self, y, t):
        S, I, R = y
        dSdt = -self.beta * S * I / self.N
        dIdt = self.beta * S * I / self.N - self.gamma * I
        dRdt = self.gamma * I
        return dSdt, dIdt, dRdt


class SIRS(MacroODE):
    def __init__(self, N, I0, R0, beta, gamma, delta):
        self.N = N
        self.I0 = I0
        self.R0 = R0
        self.S0 = N - I0 - R0
        self.initial = (self.S0, I0, R0)

        self.beta = beta
        self.gamma = gamma
        self.delta = delta
        self.reproduction_num = None    # Definition of "R_0".

    def derivative(self, y, t):
        S, I, R = y
        dSdt = -self.beta * S * I / self.N + self.delta * R
        dIdt = self.beta * S * I / self.N - self.gamma * I
        dRdt = self.gamma * I - self.delta * R
        return dSdt, dIdt, dRdt


class SEIR(MacroODE):
    def __init__(self, N, E0, I0, R0, sigma, beta, gamma):
        self.N = N
        self.E0 = E0
        self.I0 = I0
        self.R0 = R0
        self.S0 = N - E0 - I0 - R0
        self.initial = (self.S0, E0, I0, R0)

        self.sigma = sigma
        self.beta = beta
        self.gamma = gamma

    def derivative(self, y, t):
        S, E, I, R = y
        dSdt = -self.beta * S * I / self.N
        dEdt = self.beta * S * I / self.N - self.sigma * E
        dIdt = self.sigma * E - self.gamma * I
        dRdt = self.gamma * I
        return dSdt, dEdt, dIdt, dRdt


class SEIRD(MacroODE):
    def __init__(self, N, E0, I0, R0, D0, sigma, beta, gamma, p, rho):
        self.N = N
        self.E0 = E0
        self.I0 = I0
        self.R0 = R0
        self.D0 = D0
        self.S0 = N - E0 - I0 - R0 - D0
        self.initial = (self.S0, E0, I0, R0, D0)

        self.sigma = sigma
        self.beta = beta
        self.gamma = gamma
        self.p = p
        self.rho = rho

    def derivative(self, y, t):
        S, E, I, R, D = y
        dSdt = - self.beta * S * I / self.N
        dEdt = self.beta * S * I / self.N - self.sigma * E
        dIdt = self.sigma * E - (1 - self.p) * self.gamma * I \
                              - self.p * self.rho * I
        dRdt = (1 - self.p) * self.gamma * I
        dDdt = self.p * self.rho * I
        return dSdt, dEdt, dIdt, dRdt, dDdt


class EpiModel(object):
    """docstring for EpiModel"""

    def __init__(self, model, data_dir=None):
        self.model = model
        self.data_dir = data_dir

    def simulate(self, t):
        results = odeint(self.model.derivative, self.model.initial, t)
        return results.T


########################################################################


if __name__ == '__main__':
    # Seq : N, I0, beta
    si = SI(1000, 10, 0.2)
    # Seq : N, I0, beta, gamma
    sis = SIS(1000, 10, 0.2, 0.1)
    # Seq : N, I0, R0, beta, gamma
    sir = SIR(1000, 50, 0, 0.4, 0.1)
    # Seq : N, E0, I0, R0, D0, beta, gamma, delta, alpha, rho
    # seird = SEIRD(1000, 1, 0, 0, 0, )

    days = np.linspace(0, 80, 80)

    # ------------------------------------------------------------------

    # epi = EpiModel(si)
    # s, i = epi.simulate(days)

    # # Plot the data on three separate curves for S(t), I(t) and R(t)
    # fig = plt.figure(facecolor='w')
    # plt.plot(days, s / si.N, 'b', alpha=0.5, lw=2, label='Susceptible')
    # plt.plot(days, i / si.N, 'r', alpha=0.5, lw=2, label='Infected')
    # plt.xlabel('Time /days')
    # plt.ylabel('Number (1000s)')
    # plt.grid(True)
    # plt.legend()

    # plt.show()

    # ------------------------------------------------------------------

    epi = EpiModel(sir)
    s, i, r = epi.simulate(days)

    # Plot the data on three separate curves for S(t), I(t) and R(t)
    fig = plt.figure(facecolor='w')
    plt.plot(days, s / sis.N, 'b', alpha=0.5, lw=2, label='Susceptible')
    plt.plot(days, i / sis.N, 'r', alpha=0.5, lw=2, label='Infected')
    plt.plot(days, r / sis.N, 'g', alpha=0.5, lw=2, label='Recovered')
    plt.xlabel('Time /days')
    plt.ylabel('Number (1000s)')
    plt.grid(True)
    plt.legend()

    plt.show()

    # ------------------------------------------------------------------

    # epi = EpiModel(sir)
    # s, i, r = epi.simulate(days)

    # # Plot the data on three separate curves for S(t), I(t) and R(t)
    # fig = plt.figure(facecolor='w')
    # plt.plot(days, s / sir.N, 'b', alpha=0.5, lw=2, label='Susceptible')
    # plt.plot(days, i / sir.N, 'r', alpha=0.5, lw=2, label='Infected')
    # plt.plot(days, r / sir.N, 'g', alpha=0.5, lw=2,
    #          label='Recovered with immunity')
    # plt.xlabel('Time /days')
    # plt.ylabel('Number (1000s)')
    # plt.grid(True)
    # plt.legend()

    # plt.show()



########################################################################

'''

Reference:
+ scipy求解微分方程: https://www.twblogs.net/a/5b7afb882b7177539c248a16

Author: Kuo Chun-Lin
Date  : 2020.06.06 - 2020.08.11
'''

########################################################################

import copy
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

from scipy.integrate import odeint

########################################################################


class NetworkODE(object):
    # The 2 necessary parameters that are easily being ignored.
    A = None
    N = None
    initial = None

    def derivative(self, z, t):
        assert self.A is not None, 'A is not defined.'
        assert self.N is not None, 'N is not defined.'
        assert self.initial is not None, 'Nodes are not initiated.'
        raise NotImplementedError()

    @property
    def model_name(self):
        return self.__class__.__name__

    @property
    def degree_base(self):
        assert self.N is not None, 'N is not defined.'
        assert self.A is not None, 'A is not defined.'
        vec = np.sum(self.A, axis=0)
        return self.N * vec / np.sum(vec)

    @property
    def acq_base(self):
        assert self.N is not None, 'N is not defined.'
        assert self.A is not None, 'A is not defined.'
        vec = np.sum(self.A.sum(axis=0) * self.A, axis=1)
        return self.N * vec / np.sum(vec)

    def first_nei(self, state):
        assert self.N is not None, 'N is not defined.'
        assert self.A is not None, 'A is not defined.'
        state = self.N * state / np.sum(state)
        vec = np.sum(self.A, axis=0) * state
        return self.N * vec / np.sum(vec)

    def second_nei(self, state):
        assert self.N is not None, 'N is not defined.'
        assert self.A is not None, 'A is not defined.'
        state = self.N * state / np.sum(state)
        vec = np.sum(self.A.sum(axis=0) * self.A, axis=1) * state
        return self.N * vec / np.sum(vec)

    def inv_idx(self, i, total):
        idx = np.ones(total, dtype=np.bool)
        idx[i] = 0
        return idx


class SI(NetworkODE):
    def __init__(self, A, I0, beta):
        # numpy 2D Adjacent matrix
        self.A = A
        self.N = len(A)

        # Randomly assign the non-repeated infected and recovered nodes.
        idx = np.random.choice(np.arange(self.N), I0, replace=False)
        self.I0 = np.zeros((self.N,))
        self.I0[idx] = 1

        # Init matrix should be stacked into a 1D array.
        self.initial = np.hstack([1 - self.I0, self.I0])
        self.beta = beta

    def derivative(self, z, t):
        # The initial "z" starts from "self.initial".
        b = self.beta * z[0:self.N] * np.dot(self.A, z[self.N:2 * self.N])
        return np.hstack([-b, b])


class SIS(NetworkODE):
    def __init__(self, A, I0, beta, gamma):
        # numpy 2D Adjacent matrix
        self.A = A
        self.N = len(A)

        # Randomly assign the non-repeated infected and recovered nodes.
        idx = np.random.choice(np.arange(self.N), I0, replace=False)
        self.I0 = np.zeros((self.N,))
        self.I0[idx] = 1

        # Init state array is the combination of "s(t)" and "x(t)".
        self.initial = np.hstack([1 - self.I0, self.I0])

        self.beta = beta
        self.gamma = gamma
        self.reproduction_num = beta / gamma    # Definition of "R_0".

    def derivative(self, z, t):
        # The initial "z" starts from "self.initial".
        b = self.beta * z[0:self.N] * np.dot(self.A, z[self.N:2 * self.N])
        r = self.gamma * z[self.N:2 * self.N]
        return np.concatenate([-b + r, b - r])


class SIR(NetworkODE):
    def __init__(self, A, I0, R0, beta, gamma):
        # numpy 2D Adjacent matrix
        self.A = A
        self.N = len(A)

        # Randomly assign the non-repeated infected and recovered nodes.
        idx = np.random.choice(np.arange(self.N), I0 + R0, replace=False)
        self.I0 = np.zeros((self.N,))
        self.R0 = np.zeros((self.N,))
        self.I0[idx[:I0]] = 1
        self.R0[idx[I0:I0 + R0]] = 1

        # Init matrix should be stacked into a 1D array.
        self.initial = np.hstack([1 - self.I0 - self.R0,    # s(t)
                                  self.I0,                  # i(t)
                                  self.R0])                 # r(t)
        self.beta = beta
        self.gamma = gamma
        self.reproduction_num = beta / gamma    # Definition of "R_0".

    def derivative(self, z, t):
        # The initial "z" starts from "self.initial".
        b = self.beta * z[0:self.N] * np.dot(self.A, z[self.N:2 * self.N])
        r = self.gamma * z[self.N:2 * self.N]
        return np.hstack([-b, b - r, r])


class SIRV(NetworkODE):
    """docstring for SIRV"""

    def __init__(self, A, I0, R0, V0, beta, gamma, eps):
        # numpy 2D Adjacent matrix
        self.A = A
        self.N = len(A)

        # Randomly assign the non-repeated infected, recovered, and vacci nodes.
        idx = np.random.choice(np.arange(self.N), I0 + R0 + V0, replace=False)
        self.I0 = np.zeros((self.N,))
        self.R0 = np.zeros((self.N,))
        self.V0 = np.zeros((self.N,))
        self.I0[idx[:I0]] = 1
        self.R0[idx[I0:I0 + R0]] = 1
        self.V0[idx[I0 + R0:I0 + R0 + V0]] = 1

        # Init matrix should be stacked into a 1D array.
        self.initial = np.hstack([1 - self.I0 - self.R0 - self.V0,
                                  self.I0, self.R0, self.V0])
        self.beta = beta
        self.gamma = gamma
        self.eps = eps
        self.reproduction_num = beta / gamma

    def derivative(self, z, t):
        Pk = 1    # default.
        # Pk = self.degree_base
        # Pk = self.acq_base
        # Pk = self.first_nei(z[self.N * 1:2 * self.N])
        # Pk = self.second_nei(z[self.N * 1:2 * self.N])

        b = self.beta * z[0:self.N] * np.dot(self.A, z[self.N:2 * self.N])
        r = self.gamma * z[self.N:2 * self.N]
        v = (1 - self.beta) * self.eps * z[0:self.N] * Pk
        return np.hstack([-b - v, b - r, r, v])


class SIRVir1(NetworkODE):
    def __init__(self, A, I0, R0, V0, i0, r0, beta1, gamma1, alpha1,
                 beta2, gamma2, alpha2, eps):
        # numpy 2D Adjacent matrix
        self.A = A
        self.N = len(A)

        idx = np.random.choice(np.arange(self.N),
                               I0 + R0 + V0 + i0 + r0,
                               replace=False)
        self.I0 = np.zeros((self.N,))
        self.R0 = np.zeros((self.N,))
        self.V0 = np.zeros((self.N,))
        self.i0 = np.zeros((self.N,))
        self.r0 = np.zeros((self.N,))
        self.I0[idx[:I0]] = 1
        self.R0[idx[I0:I0 + R0]] = 1
        self.V0[idx[I0 + R0:I0 + R0 + V0]] = 1
        self.i0[idx[I0 + R0 + V0:I0 + R0 + V0 + i0]] = 1
        self.r0[idx[I0 + R0 + V0 + i0:I0 + R0 + V0 + i0 + r0]] = 1

        self.initial = np.hstack(
            [1 - self.I0 - self.R0 - self.V0 - self.i0 - self.r0,
             self.I0, self.R0, self.V0, self.i0, self.r0]
        )

        self.beta1 = beta1
        self.beta2 = beta2
        self.gamma1 = gamma1
        self.gamma2 = gamma2
        self.alpha1 = alpha1
        self.alpha2 = alpha2
        self.eps = eps

    def derivative(self, z, t):
        Pk = 1    # default.

        # # Degree-based Target Immunization
        # Pk = np.zeros(self.N)
        # thres = 0.1
        # Pk[self.get_nodes(thres)] = 1

        # # Acquaintance Immunization
        # Pk = np.zeros(self.N)
        # thres = 0.1
        # Pk[self.get_neighbors(thres)] = 1

        b1 = self.beta1 * z[:1 * self.N] * \
            np.dot(self.A, z[self.N * 1:2 * self.N])
        r1 = self.gamma1 * z[self.N * 1:2 * self.N]
        a1 = self.alpha1 * z[self.N * 2:3 * self.N]
        v = self.eps * z[:1 * self.N] * Pk

        b2 = self.beta2 * z[self.N * 3:4 * self.N] * \
            np.dot(self.A, z[self.N * 1:2 * self.N])
        r2 = self.gamma2 * z[self.N * 4:5 * self.N]
        a2 = self.alpha2 * z[self.N * 5:]
        return np.hstack([-b1 - v + a1 + a2,
                          b1 - r1,
                          r1 - a1,
                          v - b2,
                          b2 - r2,
                          r2 - a2])


class SIRVir2(NetworkODE):
    def __init__(self, A, I0, R0, V0, i0, r0, beta1, gamma1, alpha1,
                 beta2, gamma2, alpha2, eps, eps1v, eps2v):
        # numpy 2D Adjacent matrix
        self.A = A
        self.N = len(A)

        idx = np.random.choice(np.arange(self.N),
                               I0 + R0 + V0 + i0 + r0,
                               replace=False)
        self.I0 = np.zeros((self.N,))
        self.R0 = np.zeros((self.N,))
        self.V0 = np.zeros((self.N,))
        self.i0 = np.zeros((self.N,))
        self.r0 = np.zeros((self.N,))
        self.I0[idx[:I0]] = 1
        self.R0[idx[I0:I0 + R0]] = 1
        self.V0[idx[I0 + R0:I0 + R0 + V0]] = 1
        self.i0[idx[I0 + R0 + V0:I0 + R0 + V0 + i0]] = 1
        self.r0[idx[I0 + R0 + V0 + i0:I0 + R0 + V0 + i0 + r0]] = 1

        self.initial = np.hstack(
            [1 - self.I0 - self.R0 - self.V0 - self.i0 - self.r0,
             self.I0, self.R0, self.V0, self.i0, self.r0]
        )

        self.beta1 = beta1
        self.beta2 = beta2
        self.gamma1 = gamma1
        self.gamma2 = gamma2
        self.alpha1 = alpha1
        self.alpha2 = alpha2
        self.eps = eps
        self.eps1v = eps1v
        self.eps2v = eps2v

    def derivative(self, z, t):
        Pk = 1    # default.
        P1k = 1
        P2k = 1

        b1 = self.beta1 * z[:1 * self.N] * \
            np.dot(self.A, z[self.N * 1:2 * self.N])
        r1 = self.gamma1 * z[self.N * 1:2 * self.N]
        a1 = self.alpha1 * z[self.N * 2:3 * self.N]
        e1 = self.eps1v * z[self.N * 2:3 * self.N] * P1k
        v = self.eps * z[:1 * self.N] * Pk

        b2 = self.beta2 * z[self.N * 3:4 * self.N] * \
            np.dot(self.A, z[self.N * 1:2 * self.N])
        r2 = self.gamma2 * z[self.N * 4:5 * self.N]
        a2 = self.alpha2 * z[self.N * 5:]
        e2 = self.eps2v * z[self.N * 5:] * P2k
        return np.hstack([-b1 - v + a1 + a2,
                          b1 - r1,
                          r1 - a1 - e1,
                          v - b2 + e1 + e2,
                          b2 - r2,
                          r2 - a2 - e2])


class SEINRVeinr(NetworkODE):
    def __init__(self, A, E0, I0, N0, R0, V0, e0, i0, n0, r0,
                 beta1, theta1, eta1, gamma1, pi1, alpha1, eps,
                 beta2, theta2, eta2, gamma2, pi2, alpha2):
        # numpy 2D Adjacent matrix
        self.A = A
        self.N = len(A)

        idx = np.random.choice(np.arange(self.N),
                               I0 + R0 + V0 + i0 + r0,
                               replace=False)
        self.E0 = np.zeros((self.N,))
        self.I0 = np.zeros((self.N,))
        self.N0 = np.zeros((self.N,))
        self.R0 = np.zeros((self.N,))
        self.V0 = np.zeros((self.N,))
        self.e0 = np.zeros((self.N,))
        self.i0 = np.zeros((self.N,))
        self.n0 = np.zeros((self.N,))
        self.r0 = np.zeros((self.N,))

        num = 0
        self.E0[idx[num:num + E0]] = 1
        num += E0
        self.I0[idx[num:num + I0]] = 1
        num += I0
        self.N0[idx[num:num + N0]] = 1
        num += N0
        self.R0[idx[num:num + R0]] = 1
        num += R0
        self.V0[idx[num:num + V0]] = 1
        num += V0
        self.e0[idx[num:num + e0]] = 1
        num += e0
        self.i0[idx[num:num + i0]] = 1
        num += i0
        self.n0[idx[num:num + n0]] = 1
        num += n0
        self.r0[idx[num:num + r0]] = 1

        self.initial = np.hstack(
            [1 - self.E0 - self.I0 - self.N0 - self.R0 - self.V0 -
             self.e0 - self.i0 - self.n0 - self.r0,
             self.E0, self.I0, self.N0, self.R0, self.V0,
             self.e0, self.i0, self.n0, self.r0]
        )

        self.beta1, self.beta2 = beta1, beta2
        self.theta1, self.theta2 = theta1, theta2
        self.eta1, self.eta2 = eta1, eta2
        self.gamma1, self.gamma2 = gamma1, gamma2
        self.pi1, self.pi2 = pi1, pi2
        self.alpha1, self.alpha2 = alpha1, alpha2
        self.eps = eps

    def derivative(self, z, t):
        # 0, 1, 2, 3, 4, 5, 6, 7, 8, 9
        # S, E, I, N, R, V, e, i, n, r
        Pk = 1    # default.

        # Degree-based Target Immunization
        Pk = np.zeros(self.N)
        thres = 10
        Pk[self.get_nodes(thres)] = 1

        # # Acquaintance Immunization
        # Pk = np.zeros(self.N)
        # thres = 0.1
        # Pk[self.get_neighbors(thres)] = 1

        b1 = self.beta1 * z[:1 * self.N] * \
            np.dot(self.A, z[self.N * 2:3 * self.N])
        th1 = self.theta1 * z[self.N * 1:2 * self.N]
        e1 = (1 - self.theta1) * self.eta1 * z[self.N * 1:2 * self.N]
        r1 = self.gamma1 * z[self.N * 2:3 * self.N]
        p1 = self.pi1 * z[self.N * 3:4 * self.N]
        a1 = self.alpha1 * z[self.N * 4:5 * self.N]
        v = (1 - self.beta1) * self.eps * z[:1 * self.N] * Pk

        b2 = self.beta2 * z[self.N * 5:6 * self.N] * \
            np.dot(self.A, z[self.N * 2:3 * self.N])
        th2 = self.theta2 * z[self.N * 6:7 * self.N]
        e2 = (1 - self.theta2) * self.eta2 * z[self.N * 6:7 * self.N]
        r2 = self.gamma2 * z[self.N * 7:8 * self.N]
        p2 = self.pi2 * z[self.N * 8:9 * self.N]
        a2 = self.alpha2 * z[self.N * 9:]
        return np.hstack([-b1 - v + a1 + a2,  # S
                          b1 - th1 - e1,      # E
                          th1 - r1,           # I
                          e1 - p1,            # N
                          r1 + p1 - a1,       # R
                          v - b2,             # V
                          b2 - th2 - e2,      # e
                          th2 - r2,           # i
                          e2 - p2,            # n
                          r2 + p2 - a2])      # r


class SIRS(NetworkODE):
    def __init__(self, A, I0, R0, beta, gamma, delta):
        # numpy 2D Adjacent matrix
        self.A = A
        self.N = len(A)

    def derivative(self, z, t):
        pass


class SEIR(NetworkODE):
    def __init__(self, A, E0, I0, R0, sigma, beta, gamma):
        # numpy 2D Adjacent matrix
        self.A = A
        self.N = len(A)

    def derivative(self, z, t):
        pass


class SEIRD(NetworkODE):
    def __init__(self, A, E0, I0, R0, D0, sigma, beta, gamma, p, rho):
        # numpy 2D Adjacent matrix
        self.A = A
        self.N = len(A)

    def derivative(self, z, t):
        pass


########################################################################


class EpiModel(object):
    """ Take the instance of a Epidemic model as the input for a
    simulation within a period. """

    def __init__(self, G, eq, num_state, params, data_dir=None,
                 state_colors=['blue', 'red', 'green']):
        self.G = G
        self.A = np.array(nx.adj_matrix(G).todense())
        self.N = len(self.A)
        self.model = eq(self.A, *params)

        self.num_state = num_state
        self.data_dir = data_dir

        assert self.model.initial is not None, 'Initial vector is not defined.'

    def simulate(self, t):
        # Put the model into a differential equation solver.
        results = odeint(self.model.derivative, self.model.initial, t)
        # Split the results into a proper number of part.
        return np.split(results, self.num_state, axis=1)


########################################################################


if __name__ == '__main__':
    # Seq: (The number of nodes,
    #       The number of random edges to add for each new node,
    #       Probability of adding a triangle after adding a random edge)
    G_ws = nx.watts_strogatz_graph(40, 5, 0.4)
    # G_pl = nx.powerlaw_cluster_graph(50, 5, 0.4)
    A = np.array(nx.adj_matrix(G_ws).todense())
    days = np.arange(0, 10, 0.1)

    # ------------------------------------------------------------------

    # # Seq : Graph, I0, beta
    # si = SI(A, 4, 0.6)
    # # Seq : Graph, I0, beta, gamma
    # sis = SIS(A, 4, 0.6, 0.1)
    # # Seq : Graph, I0, R0, beta, gamma
    # sir = SIR(A, 4, 2, 0.4, 0.6)
    # # Seq: Graph, I0, R0, V0, beta, gamma, eps
    # sirv = SIRV(A, 4, 2, 2, 0.6, 0.2, 0.05)

    # model = sir
    # num_state = 3

    # simulate = odeint(model.derivative, model.initial, days)
    # s, i, r = np.array(np.split(simulate, num_state, axis=1))

    # ------------------------------------------------------------------

    epi = EpiModel(G_ws, SIR, num_state=3, params=[4, 2, 0.4, 0.6])
    s, i, r = epi.simulate(days)
    # epi.step(1)

    # Plot the data on three separate curves for S(t) and I(t).
    fig = plt.figure(facecolor='w')
    plt.plot(s, color='blue')
    plt.plot(i, color='red')
    plt.plot(r, color='green')
    # plt.plot(v, color='orange')
    plt.xlabel('Time unit', fontsize=15)
    plt.ylabel('Probability (%)', fontsize=15)
    plt.xlim((0, 100))
    plt.ylim((0, 1))
    plt.grid(True)
    plt.legend()
    plt.show()




# https://medium.com/@joel.barmettler/how-to-upload-your-python-package-to-pypi-65edc5fe9c56

import pathlib
from setuptools import setup, find_packages

__author__ = 'Guo Jun-Lin'
__license__ = "BSD 2 Clause"
__email__ = "guojl19@mails.tsinghua.edu.cn"

# Get the long description from the README file
# with open(path.join(here, 'README.md'), encoding='utf-8') as f:
#    long_description = f.read()

# The directory containing this file.
HERE = pathlib.Path(__file__).parent
# The text of the README file
README = (HERE / 'README.md').read_text()

install_requires = ['networkx', 'numpy', 'matplotlib',
                    'tqdm', 'scipy', 'cdlib', 'future']

keywords = ['epidemic', 'vaccination', 'graph', 'networkx', 'community',
            'propagation', 'differential-equation', 'covid-19', 'cdlib']

setup(name='epidemix',
      version='1.1.2',
      license='BSD-2-Clause',
      description='Simulation of Epidemic Propagation on a Network',
      long_description=README,
      long_description_content_type='text/markdown',
      url='https://github.com/khle08/epidemix',
      author='Guo Jun-Lin',
      author_email='guojl19@mails.tsinghua.edu.cn',
      classifiers=[
          # How mature is this project? Common values are
          #   3 - Alpha
          #   4 - Beta
          #   5 - Production/Stable
          'Development Status :: 3 - Alpha',

          # Indicate who your project is intended for
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Build Tools',

          # Pick your license as you wish (should match "license" above)
          'License :: OSI Approved :: BSD License',

          "Operating System :: OS Independent",

          # Specify the Python versions you support here. In particular, ensure
          # that you indicate whether you support Python 2, Python 3 or both.
          # 'Programming Language :: Python',
          # 'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3'
      ],
      keywords=keywords,
      install_requires=install_requires,
      packages=['epidemix'],
      include_package_data=True,
      entry_points={'console_scripts': ['epidemix=epidemix.__main__:main']}
      )



########################################################################

'''

Reference:
+ scipy求解微分方程: https://www.twblogs.net/a/5b7afb882b7177539c248a16

Author: Kuo Chun-Lin
Date  : 2020.06.06 - 2020.06.17
'''

########################################################################

import copy
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

from scipy.integrate import odeint

########################################################################


class EpiModel(object):
    """ Take the instance of a Epidemic model as the input for a
    simulation within a period. """

    def __init__(self, G, eq, num_state, params, data_dir=None,
                 state_colors=['blue', 'red', 'green', 'orange']):
        self.G = G
        self.A = np.array(nx.adj_matrix(G).todense())
        self.N = len(self.A)
        self.model = eq(self.A, *params)
        # Get the node position so that the drawn graph can be fixed.
        self.pos = nx.spring_layout(self.G)
        self.num_state = num_state
        self.state_queue = np.arange(num_state)
        self.state_colors = np.array(state_colors)

        # Get the initial states of all nodes.
        self.initial = np.array(np.split(self.model.initial, num_state))
        # Create an array to record the state transition.
        states = np.zeros(self.N)
        # Assign each node to ...
        for i, state in enumerate(self.initial):
            # ... diff state represented by a number.
            states[state.astype(np.bool)] = i

        # Summarize the state and color of each node to an independent arr.
        self.state_list = self.state_queue[states.astype(np.int)]
        self.color_list = self.state_colors[states.astype(np.int)]

        # Assign the initial information to the node attribute.
        self.G = self.set_graph(self.state_list,
                                self.color_list,
                                G=self.G)

        self.model_name = self.model.model_name
        self.data_dir = data_dir
        self.sim = None
        self.sims = None
        self.vacci_num = 0
        self.vacci_sum = 0
        self.vacci_list = []
        self.vacci_func = None
        self.t = None
        self.propagating = []

        assert self.model.initial is not None, '[!] Initial vector is not defined.'

    def set_graph(self, state_list, color_list, G):
        # This func take a graph as input and append attributes accordingly.
        G_ = copy.deepcopy(G)
        states_ = zip(np.arange(len(state_list)), state_list)
        colors_ = zip(np.arange(len(color_list)), color_list)
        nx.set_node_attributes(G_, dict(states_), 'state')
        nx.set_node_attributes(G_, dict(colors_), 'color')
        return G_

    def select_nodes(self, A, state=0, nei_thres=0):
        # Determine those nodes in "0" state.
        opt_1 = np.where(self.state_list == state)[0]
        # Determine those nodes with neighbors.
        opt_2 = np.where(np.sum(A, axis=1) > nei_thres)[0]
        # Find only those "0" states while also connecting with the others.
        return np.intersect1d(opt_1, opt_2)

    def select_neighbors(self, A, state, nei_state, nei_thres=0):
        # Determine which nodes to pick up neighbors.
        opt = self.select_nodes(A, state)
        # Count the total neighbor numbers of the selected nodes.
        nei_nums = np.sum(A[opt, :], axis=0)
        # Determine which nodes are the neighbors of the selected nodes.
        idx_nei = np.where(nei_nums > nei_thres)[0]
        # Find out the nodes with selected state.
        idx_in = np.where(self.state_list == nei_state)[0]
        return np.intersect1d(idx_nei, idx_in)

    def disconnect_links(self, state, ratio=1):
        opt = self.select_nodes(self.A, state=3, nei_thres=0)
        # Randomly pick up certain selected nodes.
        idx = np.random.choice(opt, round(len(opt) * ratio), replace=False)
        self.A[idx, :] = 0
        self.A[:, idx] = 0

        self.G = self.set_graph(self.state_list,
                                self.color_list,
                                G=nx.Graph(self.A))

    def simulate(self, t, param=None):
        # Only the first simulation will assign self.t attribute.
        if self.t is None:
            self.t = t

        # Change the adjacent matrix using a pre-defined strategy.
        if param is not None:
            assert self.vacci_func is not None, '[!] Vacci func is not defined.'
            self.A = self.vacci_func(self.A, param)
            # Update graph if the A matrix is modified.
            self.G = self.set_graph(self.state_list,
                                    self.color_list,
                                    G=nx.Graph(self.A))

            self.vacci_sum += self.vacci_num

        # Put the model into a differential equation solver.
        self.sim = odeint(self.model.derivative, self.model.initial, t)
        # Split the results into a proper number of part.
        self.sims = np.array(np.split(self.sim, self.num_state, axis=1))
        # shape 3D: (num_state, num_period, num_node)
        return self.sims

    def reconstruct(self, t, idx, param=None):
        assert self.sims is not None, '[!] simulation is not done yet.'
        # Take a copy of "self.sim" and find the idx-th row as ...
        sim_ori = copy.deepcopy(self.sim)
        # ... the new initial probability of each node.
        self.model.initial = sim_ori[idx]

        # Change the adjacent matrix using a pre-defined strategy.
        if param is not None:
            assert self.vacci_func is not None, '[!] Vacci func is not defined.'
            self.A = self.vacci_func(self.A, param)
            # Update graph if the A matrix is modified.
            self.G = self.set_graph(self.state_list,
                                    self.color_list,
                                    G=nx.Graph(self.A))

            self.vacci_sum += self.vacci_num

        # Simulate the later part after the "A" matrix is changed.
        self.simulate(t[idx:])
        # Assign the new "A" sim result to the later part of the matrix.
        sim_ori[idx:] = self.sim
        # Let the new combined "sim_ori" as the self.sim (same as "simulate" func).
        self.sim = sim_ori
        # Split the results into a proper number of part.
        self.sims = np.array(np.split(self.sim, self.num_state, axis=1))
        # shape 3D: (num_state, num_period, num_node)
        return self.sims

    def count_neighbors(self, idx_nodes, target_state=1):
        # Create an array to record the number of targeted neighbors.
        nei_nums = np.zeros(len(idx_nodes))

        for i, nd in enumerate(idx_nodes):
            # Determine the neighbors of each selected node iteratively.
            neighbors = list(self.G.neighbors(nd))
            count = 0
            # Examine the state of those neighbor nodes.
            for nei in neighbors:
                # If the neighbor is an targeted neighbor ...
                if self.G.nodes[nei]['state'] == target_state:
                    # ... account the neighbor.
                    count += 1

            # Record the total targeted neighbors to the array.
            nei_nums[i] = count

        return nei_nums

    def state_transition(self, idx_nodes, probs, ori_s, offset):
        # Only when the probability of next state is larger than ...
        # ... a uniformly generated random number, the node state ...
        # ... will be changed.
        idx_trans = idx_nodes[probs > np.random.rand(len(probs))]
        self.state_list[idx_trans] = self.state_queue[int(ori_s + offset)]
        self.color_list[idx_trans] = self.state_colors[int(ori_s + offset)]

    def __propagate(self, idx, state_list, from_state, to_state,
                    neighbor=None, update_state=False):
        if update_state:
            # If the next state is not unique, state list should be updated.
            idx_nodes = np.where(self.state_list == from_state)[0]
        else:
            # Get the indices of those nodes at current state.
            idx_nodes = np.where(state_list == from_state)[0]

        if neighbor is not None:
            # Count how many infected neighbor nodes.
            nei_nums = self.count_neighbors(idx_nodes, target_state=neighbor)
            # Pick up only those nodes with infected neighbors.
            idx_nodes = idx_nodes[nei_nums > 0]

        # Get the probabilities of each node from the sim results.
        probs = self.sims[to_state, idx, idx_nodes]
        self.state_transition(idx_nodes, probs, from_state,
                              to_state - from_state)

    def set_propagate(self, from_state, to_state, neighbor=None, update_state=False):
        self.propagating.append([from_state, to_state,
                                 neighbor, update_state])

    def step(self, idx):
        assert self.sims is not None, '[!] simulation is not done yet.'
        # Things that changed in each "step" are only the attributes of nodes.

        # Record the original state so that the updated state will not be mixed.
        state_list = copy.deepcopy(self.state_list)

        for info in self.propagating:
            self.__propagate(idx, state_list, *info)

        self.G = self.set_graph(self.state_list,
                                self.color_list,
                                G=self.G)

    def propagation(self, reconstruct_index=[], reconstruct_param=None,
                    disconnect_index=[], disconnect_state=3):
        status = []
        probs = None

        for idx in range(len(self.t)):
            record = {}
            self.step(idx)

            if idx in reconstruct_index:
                probs = self.reconstruct(self.t, idx, param=reconstruct_param)

            if idx in disconnect_index:
                self.disconnect_links(state=disconnect_state)

            self.vacci_list.append(self.vacci_sum)

            record['iteration'] = idx
            record['state'] = copy.deepcopy(self.state_list)
            record['color'] = copy.deepcopy(self.color_list)
            record['adj'] = copy.deepcopy(self.A)
            record['graph'] = copy.deepcopy(self.G)
            record['vacci_sum'] = self.vacci_sum
            status.append(record)

        return status, probs

    def visualize(self, status, index, figsize, n_row, n_col, save=False):
        plt.figure(figsize=figsize)

        for n, idx in enumerate(index):
            plt.subplot(n_row, n_col, n + 1)
            plt.title('t = {}'.format(idx))

            nx.draw(status[idx]['graph'], self.pos,
                    with_labels=True, font_color='white',
                    font_size=6, node_size=60,
                    node_color=status[idx]['color'])
        if save:
            plt.savefig('nx.jpg')

########################################################################


if __name__ == '__main__':

    from equations import SI, SIS, SIR, SIRV

    # Seq: (The number of nodes,
    #       The number of random edges to add for each new node,
    #       Probability of adding a triangle after adding a random edge)
    G_ws = nx.watts_strogatz_graph(40, 5, 0.4)
    # G_pl = nx.powerlaw_cluster_graph(50, 5, 0.4)
    days = np.arange(0, 10, 0.1)

    # ------------------------------------------------------------------

    # SI   params: I0, beta
    # SIS  params: I0, beta, gamma
    # SIR  params: I0, R0, beta, gamma
    # SIRV params: I0, R0, V0, beta, gamma, eps

    # ------------------------------------------------------------------

    epi = EpiModel(G_ws, SIRV, num_state=4, params=[4, 2, 2, 0.4, 0.2, 0.3])
    s, i, r, v = epi.simulate(days)

    epi.set_propagate(0, 1, neighbor=1, update_state=False)
    epi.set_propagate(0, 3, neighbor=None, update_state=True)
    epi.set_propagate(1, 2, neighbor=None, update_state=False)

    status, _ = epi.propagation(disconnect_index=np.arange(16),
                                disconnect_state=3)
    # status, _ = epi.propagation(reconstruct_index=np.arange(16),
    #                             reconstruct_param=None)
    epi.visualize(status, np.arange(16), figsize=(15, 15), n_row=4, n_col=4)

    # Plot the data on three separate curves for S(t) and I(t).
    fig = plt.figure(facecolor='w')
    plt.plot(s, color='blue')
    plt.plot(i, color='red')
    plt.plot(r, color='green')
    plt.plot(v, color='orange')
    plt.xlabel('Time unit', fontsize=15)
    plt.ylabel('Probability (%)', fontsize=15)
    plt.xlim((0, 100))
    plt.ylim((0, 1))
    plt.grid(True)
    plt.legend()
    plt.show()



########################################################################

'''
say something ....
'''

########################################################################

'''

Reference:
+ scipy求解微分方程: https://www.twblogs.net/a/5b7afb882b7177539c248a16

Author: Kuo Chun-Lin
Date  : 2020.06.06 - 2020.06.17
'''

########################################################################

import os
import copy
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

from tqdm import tqdm
from epidemix.epidemic import EpiModel
from epidemix.vaccination import VacciModel
from epidemix.equations import SI, SIS, SIR, SIRV

from epidemix.utils.plot import draw_probs

########################################################################

# SI   params: I0, beta
# SIS  params: I0, beta, gamma
# SIR  params: I0, R0, beta, gamma
# SIRV params: I0, R0, V0, beta, gamma, eps

# ----------------------------------------------------------------------


def main():
    # Seq: (The number of nodes,
    #       The number of random edges to add for each new node,
    #       Probability of adding a triangle after adding a random edge)
    G = nx.watts_strogatz_graph(50, 5, 0.2)
    # G = nx.powerlaw_cluster_graph(50, 5, 0.2)
    # G = nx.random_lobster(10, 0.6, 0.6, 0)  # 0: seed

    days = np.arange(0, 100, 1)
    colors = ['blue', 'red', 'green', 'orange']

    # epi = EpiModel(G, SIR, num_state=3, params=[4, 2, 0.3, 0.1])
    epi = VacciModel(G, SIR, num_state=3, params=[4, 2, 0.3, 0.1],
                     vacci_strategy='random', state_colors=colors)

    probs = epi.simulate(days)  # , param=0.5)
    # s, i, r = epi.reconstruct(days, 20, param=0.5)
    # s, i, r = epi.reconstruct(days, 40, param=0.5)
    # s, i, r = epi.reconstruct(days, 60, param=0.5)

    epi.set_propagate(0, 1, neighbor=1, update_state=False)
    epi.set_propagate(1, 2, neighbor=None, update_state=False)

    # status, _ = epi.propagation()
    status, probs = epi.propagation(reconstruct_index=[3, 7, 9],
                                    reconstruct_param=0.5)

    epi.visualize(status, np.arange(16), figsize=(15, 15), n_row=4, n_col=4)
    draw_probs(probs, colors)


########################################################################


if __name__ == '__main__':

    # ------------------------------------------------------------------
    main()    # For testing the propagation process.
    # ------------------------------------------------------------------
